'''Implementation of the restriction syntax'''
__revision__ = '$Revision$'

from elementtree.ElementTree import Element
from random import Random
from re import compile as regcompile
from syslog import syslog, LOG_INFO
from time import time

class Data(object):
    '''Data is an implementation of the Restriction syntax using ElementTree for native python datatypes'''
    def __init__(self, element):
        self.element = element
        # for ease of reading
        self.attr = self.element.attrib

    def AttrMatch(self, elem1, elem2):
        '''AttrMatch checks if all fields in a particular pair of elements match with
        respect to the restriction syntax'''
        fields = elem2.attrib.keys()
        if elem1.tag != elem2.tag:
            return False
        if elem2.attrib.get('match', None) == 'extended':
            fields.remove('match')
            for field in fields:
                (operation, val) = elem2.attrib[field].split(':', 1)
                if operation[:2] == '!d':
                    operation = operation[2:]
                if operation == '=':
                    if val != elem1.attrib.get(field, None):
                        return False
                elif operation == '!=':
                    if val == elem1.attrib.get(field, None):
                        return False
                elif operation == '>':
                    try:
                        if int(val) < int(elem1.attrib[field]):
                            continue
                    except:
                        return False
                elif operation == '<':
                    try:
                        if int(val) > int(elem1.attrib[field]):
                            continue
                    except:
                        return False
                elif operation == 're':
                    try:
                        reg = regcompile(val)
                        if not reg.match(elem1.attrib[field]):
                            return False
                    except:
                        return False
                elif operation == 'range':
                    try:
                        (rmin, rmax) = [int(ival) for ival in val.split('-')]
                        i = int(elem1.attrib.get(field, None))
                        if ((i > rmin) and (rmax > i)):
                            continue
                    except ValueError:
                        pass
                    return False
                else:
                    return False
            return True
        else:
            for field in elem2.attrib.keys():
                if ((elem2.attrib[field] != elem1.attrib.get(field, None)) and (elem2.attrib[field] != '*')):
                    return False
            return True

    def Match(self, specification):
        '''Match checks if all fields and child node fields match specification provided'''
        if self.AttrMatch(self.element, specification):
            # If we make it here, then all attributes in this element have matched
            # Now to handle children
            for child in specification.getchildren():
                found = False
                # all child nodes of the spec must be matched
                for node in self.element.findall(child.tag):
                    if self.AttrMatch(node, child):
                        found = True
                if not found:
                    return False
            return True
        else:
            return False

    def FetchElement(self, element, spec):
        '''FetchElement takes an element and a spec and produces the proper subset et Element'''
        r = Element(element.tag)
        if element.text != None:
            r.text = element.text
        if spec.attrib.get('match') == 'extended':
            # we need to check for '!d' operator
            fields = spec.attrib.keys()
            fields.remove('match')
            for field in fields:
                if spec.attrib[field][:2] != '!d':
                    r.attrib[field] = element.attrib.get(field, "BUG")
        else:
            for field in spec.attrib.keys():
                r.attrib[field] = element.attrib.get(field, "BUG")
        # Match against children and call FetchElement recursively
        for child in spec.getchildren():
            for node in element.findall(child.tag):
                if self.AttrMatch(node, child):
                    c = self.FetchElement(node, child)
                    r.append(c)
        return r

    def Fetch(self, spec=None):
        '''Fetch takes a matching spec, and produces the subset element containing only data specified'''
        if spec != None:
            return self.FetchElement(self.element, spec)
        return self.element

class ID(object):
    '''Autoincrementing id generator'''
    def __init__(self):
        self.id = 0

    def get(self):
        '''Return new ID'''
        self.id += 1
        return self.id 

class RandomID(object):
    '''Somewhat randomly selected unique ID pool'''
    def __init__(self):
        self.used = []
        self.rand = Random(int(time()))

    def get(self):
        '''Return new random id'''
        id = str(self.rand.randrange(0, 2147483639)) + str(self.rand.randrange(0, 2147483639))
        while id in self.used:
            id = str(self.rand.randrange(0, 2147483639)) + str(self.rand.randrange(0, 2147483639))
        self.used.append(id)
        return id

class DataSet(object):
    '''This class provides a list store for Data objects, complete with functions that do most of the
    restriction syntax work
    rname is the container name used in responses
    ename is the element name used for specifications
    obj is the constructor to call on new instances for the datastore
    idalloc is a unique id allocator; must be paired with self.obj function signature
    log is if add/del operations should be logged.'''
    def __init__(self, rname, ename, obj, idalloc, log):
        self.data = []
        self.rname = rname
        self.ename = ename
        self.obj = obj
        self.idalloc = idalloc
        self.log = log

    def __iter__(self):
        return iter(self.data)

    def remove(self, x):
        return self.data.remove(x)

    def append(self, x):
        return self.data.append(x)
    
    def Add(self, spec, (peer, port), callback=None):
        '''Implement restriction syntax Add for component use'''
        ret = Element(self.rname)
        for espec in spec.findall(self.ename):
            if self.idalloc:
                child = self.obj(espec, self.idalloc.get())
            else:
                child = self.obj(espec)
            if self.log:
                syslog(LOG_INFO, "Got %s from %s"%(self.ename, peer))
            self.data.append(child)
            if callback:
                callback(child, spec.attrib)
            ret.append(child.Fetch(espec))
        return ret

    def Get(self, spec, (peer, port), callback=None):
        '''Implement restriction syntax Get for component use'''
        ret = Element(self.rname)
        for espec in spec.findall(self.ename):
            for item in [datum for datum in self.data if datum.Match(espec)]:
                if callback:
                    callback(item, spec.attrib)
                ret.append(item.Fetch(espec))
        return ret

    def Del(self, spec, (peer, port), callback=None):
        '''Implement restriction syntax Del for component use'''
        ret = Element(self.rname)
        for espec in spec.findall(self.ename):
            for item in [datum for datum in self.data if datum.Match(espec)]:
                self.data.remove(item)
                if self.log:
                    syslog(LOG_INFO, "Got %s from %s" % (spec.tag, peer))
                if callback:
                    callback(item, spec.attrib)
                ret.append(item.Fetch(espec))
        return ret

class SizedSet(DataSet):
    '''SizedSet is a DataSet instance that rotates through data, based on min and max'''
    def __init__(self, rname, ename, obj, idalloc, log, smin, smax):
        DataSet.__init__(self, rname, ename, obj, idalloc, log)
        self.min = smin
        self.max = smax
        
    def Add(self, spec, (peer, port), callback=None):
        '''Add a new entry, and rollever if needed'''
        ret = DataSet.Add(self, spec, (peer, port), callback)
        if len(self.data) > self.max:
            self.data = self.data[-self.min:]
        return ret

